/********
 * gtksu - a lightweight GTK gui for su (for root only)
 * (C) James Budiono 2012
 * License: GNU GPL Version 3 or later
 * Note: does not claim to be secure or correct!
 * Version 2: refactor and improve GUI
 * Version 3: add prompt
 * Version 4: drop user "nobody", setuid/setgid to original user instead
 * Version 5: option to "always allow until next boot"
 * Version 6: change option to "keep password", prompt always displayed 
 * Version 7: no more pipe - just use exit code for ipc, add functions to get username/password
 * Version 8: uses gtksudoer file for finer grained permission
 * Version 9: adds group name too
 * /etc/gtksudoer format:
 * - comments prefixed by hash (#)
 * - each line contains 4 fields, separated by blanks
 * - field 1: user
 * - field 2: group
 * - field 3: application path (partial path is okay)
 * - field 4: permission, which is one of this: root, user, nopass
 * (root = require root password, user = require user password, nopass = no password required)
 * field 1, 2 and 3 accepts wildcards (? and *)
 * Version 10: clear PATH and LD_* environments, gtksudoer now can have environment settings
 * Version 11: bugfix
 ********/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <shadow.h>
#include <crypt.h>
#include <sys/wait.h>

#include <regex.h>
#include <strings.h>
#include <fnmatch.h>

#include <gtk/gtk.h>
#include <libintl.h>
#include <alloca.h>

#include "gtksu.h" // our GUI

extern char *strfry (char *string); // _GNU_SOURCE : should be in string.h but not found?

extern char **environ;

char *root_password;
char *user_password;
char *user_name;
char **user_groups;
int numgroups;
char *random_password;
int permission;

#define FAIL 0
#define PROCEED 1
#define PROCEED_KEEP 2

#define RANDOM_PASSWORD "abcdefABCDEF123456!@#$%^"
#define DEFAULT_ENV_PATH "/usr/local/bin:/usr/bin:/usr/sbin:/bin:/sbin:/usr/X11R7/bin"
#define DEFAULT_ENV_LD_LIBRARY_PATH "/lib:/usr/lib64:/usr/X11R7/lib64"

//=== gtksudoer declarations

#define GTKSUDOER "/etc/gtksudoer"
#define GTKSUDOER_PERM 0600
//#define GTKSUDOER "test.conf"
//#define GTKSUDOER_PERM 0644

#define LINEBUF 1024

#define MAXMATCHES 5
#define COMMENT_REGEX "[[:space:]]*#.*"
#define ACCESS_REGEX ("^[[:space:]]*([^[:space:]]+)" \
				   "[[:space:]]+([^[:space:]]+)" \
				   "[[:space:]]+([^[:space:]]+)" \
				   "[[:space:]]+([^[:space:]]+)" \
				   "[[:space:]]*$")
#define ENV_REGEX "^[[:space:]]*([^[:space:]]*)=([^[:space:]]*)[[:space:]]*$"
regex_t comment_regex, access_regex, env_regex;
regmatch_t matches[MAXMATCHES];

#define S_ROOT_PERM "ROOT"
#define S_USER_PERM "USER"
#define S_NOPASS_PERM "NOPASS"

#define DEFAULT_PERM 0
#define ROOT_PERM 1
#define USER_PERM 2
#define NOPASS_PERM 3


//========= start marker file stuff - must run as root =========

#define MARKER_FILE_PREFIX "/tmp/gtksu."
char marker_filename[256];

char *get_marker_file() {
	snprintf ( marker_filename, sizeof(marker_filename), MARKER_FILE_PREFIX "%s", user_name );
	return marker_filename;
}

void create_marker_file() {
	int fd;
	char *marker_file;
	
	if ( (marker_file = get_marker_file()) ) {
		fd = creat ( marker_file, 0 );
		if ( fd != -1 ) close ( fd );
	}
}

int check_marker_file() {
	char *marker_file;
	struct stat buf;
	
	if ( ( marker_file = get_marker_file() ) &&
		 ( lstat ( marker_file, &buf ) != -1 ) &&
		 ( buf.st_uid == 0 && buf.st_gid == 0 ) )
		 return TRUE;
	return FALSE;
}

void remove_marker_file() {
	char *marker_file;
	
	if ( (marker_file = get_marker_file()) ) {	
		unlink ( marker_file );
	}
}
//========= end marker file stuff =========


//========= start password info stuff - must run as root =========

// get encrypted root password - for comparison later
int get_root_password() {
	struct passwd *pw = getpwuid(0); //root
	if (!pw) return FALSE;
		
	struct spwd *spwd = getspnam(pw->pw_name);
	if (!spwd) return FALSE;
	
	root_password = strdup(spwd->sp_pwdp);
	return TRUE;
}

// get encrypted user password - for comparison later
int get_user_password() {
	struct passwd *pw = getpwuid(getuid()); //current user
	if (!pw) return FALSE;
		
	struct spwd *spwd = getspnam(pw->pw_name);
	if (!spwd) return FALSE;
	
	user_name = strdup(pw->pw_name);
	user_password = strdup(spwd->sp_pwdp);
	return TRUE;
}

int get_user_groups() {
	int i; 
	gid_t *gid_array;
	struct group *grp;

	numgroups = getgroups (0, NULL);
	gid_array = alloca (sizeof(gid_t) * numgroups);
	user_groups = malloc (sizeof (char *) * numgroups);
	getgroups (numgroups, gid_array);
	
	for (i=0; i<numgroups; i++) {
		grp = getgrgid (gid_array[i]);
		user_groups[i] = strdup (grp->gr_name);
		//printf("%d %s\n",gid_array[i], user_groups[i]);
	}
	return TRUE;
}

//========= end password info stuff =========





//====== gtksudoer checking - must run as root  =====
//note: gtksudoer must be owned by root:root and must of mode GTKSUDOER_PERM

void init_regex () {
	regcomp (&comment_regex, COMMENT_REGEX, REG_EXTENDED);
	regcomp (&access_regex, ACCESS_REGEX, REG_EXTENDED);
	regcomp (&env_regex, ENV_REGEX, REG_EXTENDED);
}

void close_regex () {
	regfree (&comment_regex);
	regfree (&access_regex);
	regfree (&env_regex);
}

void drop_comment (char *line) {
	//printf("1 %s\n", line);
	if (regexec (&comment_regex, line, MAXMATCHES, matches, 0) == 0) {
		line [matches[0].rm_so] = '\0'; //drop it
		//printf("2 %s %d\n", line, matches[0].rm_so);
	}
}

int process_line (char *line, char *user, char *groups[], int numgroups, char *path) {
	char *userpat, *groupat, *pathpat, *permpat, *envkey, *envvalue;
	int i;

	// 1. check for access configuration
	if (regexec (&access_regex, line, MAXMATCHES, matches, 0) == 0) {
		line [matches[1].rm_eo] = 0; userpat = line+matches[1].rm_so;
		line [matches[2].rm_eo] = 0; groupat = line+matches[2].rm_so;
		line [matches[3].rm_eo] = 0; pathpat = line+matches[3].rm_so;
		line [matches[4].rm_eo] = 0; permpat = line+matches[4].rm_so;
		//printf ("%s %s %s\n", userpat, pathpat, permpat);
		
		if (fnmatch (userpat, user, 0)==0 && fnmatch (pathpat, path, 0)==0) {
			for (i=0; i<numgroups; i++) {
				if (fnmatch (groupat, groups[i],0)==0) {
					if (strcasecmp(permpat, S_ROOT_PERM)==0) return ROOT_PERM; 
					else if (strcasecmp(permpat, S_USER_PERM)==0) return USER_PERM;
					else if (strcasecmp(permpat, S_NOPASS_PERM)==0) return NOPASS_PERM;					
				}
			}
		}
		
	// 2. check for environment configuration, and modify env (set/unset) as found
	} else if (regexec (&env_regex, line, MAXMATCHES, matches, 0) == 0) {
		line [matches[1].rm_eo] = 0; envkey = line+matches[1].rm_so;
		line [matches[2].rm_eo] = 0; envvalue = line+matches[2].rm_so;
		//printf ("%s == %s\n", envkey, envvalue);
		//if (*envvalue == 0) printf ("deleting %s\n", envkey);
		if (line[matches[2].rm_so] == 0) unsetenv (&line[matches[1].rm_so]);
		else setenv (envkey, envvalue, TRUE);
	}
	return DEFAULT_PERM;
}

int process_config_file (char *config_file, char *user, char *groups[], int numgroups, char *path) {
	FILE *f;
	char line[LINEBUF];
	int perm = DEFAULT_PERM;
	struct stat buf;

	//check that file is owned by root and is the right mode, otherwise ignore it
	if ( ! ( ( lstat ( config_file, &buf ) != -1 ) &&
		   ( buf.st_uid == 0 && buf.st_gid == 0 ) &&
		   ( (buf.st_mode & 0777) == GTKSUDOER_PERM )
	)) return DEFAULT_PERM;

	//find the allowable permission for combination of user & path
	f = fopen (config_file, "r");
	if (f) {
		init_regex();			
		while (fgets (line, LINEBUF, f) ) {
			if (line [strlen(line)-1] == '\n') line[strlen(line)-1] = '\0';
			drop_comment (line);
			if (line[0] == '\0') continue;
			if ( (perm = process_line (line, user, groups, numgroups, path)) ) {
				//printf ("perm %d\n", perm);
				break;
			}
		}
		close_regex();		
		fclose (f);
	}
	return perm;
}

void setup_env() {
	char **p, *q;
	char envkey[LINEBUF];
	
	if (environ) {
		// unset anything with LD_
		for (p=environ; *p; p++) {
			if (strncasecmp (*p, "LD_", 3) == 0) {
				q = strchr (*p, '=');
				if (q && (q - *p) < sizeof(envkey)) {
					strncpy (envkey, *p, (q - *p));
					envkey [(q - *p)] = 0;
					printf ("removing %s\n",envkey);
					unsetenv (envkey);
				}
			}
		}
		// set our default env
		setenv ("LD_LIBRARY_PATH", DEFAULT_ENV_LD_LIBRARY_PATH, TRUE);
		setenv ("PATH", DEFAULT_ENV_PATH, TRUE);
	}
}

//========= end of gtksudoer checking ==============





//========= start of gtksu proper ================

// init - get encrypted root/user/random password, check config file and then fork
pid_t init(char *path) {
	random_password = strfry ( strdup(RANDOM_PASSWORD) );
	//printf ("%s\n", random_password);
	if ( ! get_root_password() ) return (pid_t)-1;
	if ( ! get_user_password() ) return (pid_t)-1;
	if ( ! get_user_groups() ) return (pid_t)-1;
	//printf("%s %s\n", user_name, user_password);
	setup_env();
	permission = process_config_file (GTKSUDOER, user_name, user_groups, numgroups, path);
	//printf ("%d\n", permission);
	return fork();
}

int validate_password (char *password) {
	char *encrypted;
	
	switch (permission) {
		case USER_PERM:
			encrypted = crypt(password, user_password);
			return (strcmp(encrypted, user_password) == 0);
		
		case DEFAULT_PERM: // default permission is the same as asking root password
		case ROOT_PERM:
			encrypted = crypt(password, root_password);
			return (strcmp(encrypted, root_password) == 0);
	}
	return 0;
}

//returns FAIL, PROCEED, or PROCEED_KEEP
int confirm_password(int argc, char *argv[]) {
    GtkBuilder *builder;
    GError     *error = NULL;
    
	GtkDialog  *error_dialog;    
    GtkDialog  *password_dialog;
    GtkEntry   *password_entry;
    GtkLabel   *cmd_label, *prompt_label, *password_label;
    GtkToggleButton *keep_checkbox;
    
    char cmd[LINEBUF];
    int  result, n;
    char *password;
    int  keep = FALSE;
    int	 allow_fake_password = FALSE;
    int  marker_file_present = check_marker_file();
    
    if (permission == NOPASS_PERM) {
		if (marker_file_present) return PROCEED_KEEP;
		else return PROCEED;
	}

	setlocale (LC_ALL, "");
	bindtextdomain ("gtksu", "/usr/share/locale");
	textdomain ("gtksu");
	
    gtk_init( &argc, &argv );
    builder = gtk_builder_new();    
    
    //if( ! gtk_builder_add_from_file( builder, GTKSU_GUI, &error ) )
    if( ! gtk_builder_add_from_string( builder, GTKSU_GUI, -1, &error ) )
    {
        g_error ( "%s", error->message );
        return( FAIL );
    }
    
    password_dialog = (GtkDialog *) gtk_builder_get_object ( builder, "password_dialog" ); 
    password_entry	= (GtkEntry *)  gtk_builder_get_object ( builder, "password" ); 
    password_label	= (GtkLabel *)  gtk_builder_get_object ( builder, "password_label" ); 
    cmd_label       = (GtkLabel *)  gtk_builder_get_object ( builder, "cmd" ); 
    prompt_label    = (GtkLabel *)  gtk_builder_get_object ( builder, "prompt" );
    keep_checkbox = (GtkToggleButton *) gtk_builder_get_object ( builder, "keep" );
    g_object_unref( G_OBJECT( builder ) );
    
    //prompt for USER password instead of ROOT if permission is user
    if (permission == USER_PERM) gtk_label_set_text ( password_label, "Enter re-confirm your password:\n" );
    
	// set prompt if it's not empty
	if (*argv[1]) gtk_label_set_text ( prompt_label, argv[1] );
	
	// if marker file present, fill password with our random password
	if ( marker_file_present ) {
		gtk_entry_set_text ( password_entry, random_password );
		gtk_toggle_button_set_active ( keep_checkbox, 1 );
		allow_fake_password = TRUE;
	}
    
    // copy command lines for display
    strcpy(cmd, ""); n = 2;
    while ( n < argc ) {
		strncat (cmd, " ", sizeof (cmd) );
		strncat (cmd, argv[n++], sizeof (cmd) );
	}
    gtk_label_set_text ( cmd_label, cmd );
	
	while (TRUE) {
		result = gtk_dialog_run ( password_dialog );
		//g_debug ( "dialog result %d\n", result );
		if (result == 1) {
			password = (char *) gtk_entry_get_text ( password_entry );
			keep = gtk_toggle_button_get_active ( keep_checkbox ); 
			//g_debug ( "password %s\n", password );
			if ( ( validate_password (password) ) || 
			     ( allow_fake_password && (strcmp(password, random_password) == 0) ) ) // make sure it is not changed
			{
				if ( keep ) return PROCEED_KEEP;
				else return PROCEED;
			}
			else {
				allow_fake_password = FALSE; // fake password is only allowed once
				error_dialog = (GtkDialog *) gtk_message_dialog_new (NULL,
												  GTK_DIALOG_DESTROY_WITH_PARENT,
												  GTK_MESSAGE_ERROR,
												  GTK_BUTTONS_CLOSE,
												  "Wrong password.");
				 gtk_window_set_title ( (GtkWindow *) error_dialog, "Error !" );
				 gtk_dialog_run ( error_dialog );
				 gtk_widget_destroy ((GtkWidget *) error_dialog);
				 gtk_widget_grab_focus ( (GtkWidget *) password_entry );
			}
		} else break;
	}	
    gtk_widget_destroy ( (GtkWidget *) password_dialog );
	return FAIL;
}

//only return when execvp fails
void run_as_root(int argc, char *argv[]) {
	//change identity to root and exec
	initgroups("root", 0);
	endgrent(); 
	setgid(0);
	setuid(0);			
	execvp(argv[2],&argv[2]);	
}

int main(int argc, char *argv[]) {
	pid_t pid;
	int   result;
	
	if (argc < 3) {
		printf ( "Usage: %s prompt command [args]\n", argv[0] );
		exit (1);
	}
		
	// get root password, check config file, clear env, then fork
	pid = init(argv[2]);
	
	// what to do after fork
	switch (pid) {
		case (pid_t)-1: //error
			printf ( "Some error happens.\n" );
			exit (1); break;
			
		case 0:	
			// child - drop back to our original user so that GTK does not complain,
			// then continue to validate the password - and return our result to parent via exit code
			setgid (getgid());
			setuid (getuid());
			exit ( confirm_password(argc, argv) );

		default:
			// parent - wait for authentication result from child
			waitpid (pid, &result, 0);
			switch ( WEXITSTATUS(result) ) {
				case PROCEED:
					remove_marker_file();
					run_as_root(argc, argv); //should not return
					break;
					
				case PROCEED_KEEP:
					create_marker_file();
					run_as_root(argc, argv); //should not return
					break;
			}
			exit (1);	// fail
	}
}

